
#################################################################################
# Groovix pacauth pam module to be used with pam_python.so
# pam_python home page: http://ace-host.stuart.id.au/russell/files/pam_python/
#################################################################################

########################################################################
# Copyright 2011  Open Sense Solutions LLC  http://open-sense.com
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.  You should have received a copy of the GNU General Public License along with this program.  If not, see http://www.gnu.org/licenses/
########################################################################


import re
from sys import path
import os.path

#example usage in /etc/pam.d/<program_name>
# auth     requisite       pam_python.so /lib/security/pam_pacauth.py [passwordprompt=PIN:]


path.append('/usr/share/pacauth/src')

import pacauth

def pam_sm_authenticate(pamh, flags, argv):
    debug=0
    rc=pamh.PAM_SUCCESS
    rm="OK"

    submitted_username=pamh.get_user()
    if debug:
      print submitted_username


    msg_style=pamh.PAM_PROMPT_ECHO_OFF

# check for /etc/groovix/nologin
   
    if os.path.isfile('/etc/groovix/nologin') and not ( ( submitted_username == 'gvuser' ) or ( submitted_username == 'staff' ) )  :
        rc=pamh.PAM_AUTH_ERR
        msg_style=pamh.PAM_ERROR_MSG;
        msg='Logins are not allowed at this time.'
        prompt=pamh.Message(msg_style, msg)
        response=pamh.conversation(prompt)
        return rc



    #check for passwordless login
    nopinrequired=0
    
    try:
      pwlfile = open("/etc//groovix/passwordlesslogins", "r")
      for line in pwlfile:
       if ( line.rstrip() == submitted_username ):
         if debug:
            print "passwordless login match"
         nopinrequired=1;
         break
    except Exception, err:
      if debug:
        print str(err)
      nopinrequired=0;

    # if this is for xlock, we force them to type in their pin since passwordless screen locking makes no sense
    if "xlock" in argv :
        nopinrequired=0

    if (nopinrequired==1):

       submitted_pin="nopinrequired"
    else:
       #msg="PIN : "
       msg=getPwPromptArg(argv)
       prompt=pamh.Message(msg_style, msg)
       response=pamh.conversation(prompt)

       submitted_pin=response.resp

       if debug:
           print response.resp
           print response.resp_retcode

    sanitized_user=submitted_username.lower()
    sanitized_user=sanitized_user.replace(' ','')

    if debug:
        print type(sanitized_user)

    #add a u at the beginning if the username starts with a number
    if re.match('[0-9]',sanitized_user):
        sanitized_user='u'+sanitized_user


    #check username and pin for invalid characters

    #do we allow blank username and/or pin? sure, pass it on to pacauth and let us deal with that there if we need special cases

    if re.match('^[a-zA-Z0-9\_\.\-\@]*$',sanitized_user) and re.match('^[a-zA-Z0-9\_\.\-\@\~\!\#\$\%\^\&\*\(\)]*$',submitted_pin):


        pamh.user=sanitized_user
        if debug:
            print pamh.get_user()
        
        try:
           if "xlock" in argv :
              #then we just need to do the check and not full login prep
              rc,rm=pacauth.run(sanitized_user,submitted_pin,False)
           else:
              rc,rm=pacauth.run(sanitized_user,submitted_pin,True)
        except:
           print "exception"
           print str(formatExceptionInfo())
        if debug:
            print "rc is ",rc
        if rc==0 or rc==2:
            rc=pamh.PAM_SUCCESS 
        else:
            rc=pamh.PAM_AUTH_ERR
        if debug:
            print "rc is ",rc
        

    else:
        rm="invalid barcode or pin"
        rc=pamh.PAM_AUTH_ERR

    
    if rc!=pamh.PAM_SUCCESS :
        msg_style=pamh.PAM_ERROR_MSG;
        msg=rm
        prompt=pamh.Message(msg_style, rm)
        response=pamh.conversation(prompt)

    return rc

def pam_sm_setcred(pamh, flags, argv):
  return pamh.PAM_SUCCESS


def formatExceptionInfo(maxTBlevel=4):
     import traceback 
     import sys
     cla, exc, trbk = sys.exc_info()
     excName = cla.__name__
     try:
         excArgs = exc.__dict__["args"]
     except KeyError:
         excArgs = "<no args>"
     excTb = traceback.format_tb(trbk, maxTBlevel)
     return (excName, excArgs, exc, excTb)

def getPwPromptArg(argv):
    #default prompt:
    pw_prompt="PIN:"
    for arg in argv[1:]:
      if arg.startswith("passwordprompt="):
         pw_prompt= arg[15:]
  
    return pw_prompt



