########################################################################
# Groovix PacAuth Customizable Authorization functions
########################################################################

########################################################################
# Copyright 2007-2011  Open Sense Solutions LLC  http://open-sense.com
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.  You should have received a copy of the GNU General Public License along with this program.  If not, see http://www.gnu.org/licenses/
########################################################################

import ConfigParser

import shlex

import subprocess

import re

import syslog

from datetime import datetime

from httplib2 import Http


##############################################
#change to a different auth type as necessary
##############################################
def my_auth(myname,mypass):
   return sip_auth(myname,mypass)

####################################################################
# default authentication routines follow, customize as necessary
# when defining new auth methods be sure to check if username is applicable
# and that the auth method is fully configured so that we don't mistake
# an undefined server for a temporarily down connection and let people
# login without actually verifying usernames/passwords
####################################################################


####################################################
#SIP AUTHENTICATION 
####################################################

def sip_auth(myusername,mypassword):

  #take off the beginning u for custom auth, everything else use myunixname
  #match=re.search(r'^u(\d{4,})$',username)
  #allow dashes in barcodes
  match=re.search(r'^u(\d+\-*\d+)$',myusername)
  #unless we have compatible name don't even try custom auth
  if (not match):
    return (33,"sip authentication not applicable for this username")
  mysipname=match.group(1)

  config=ConfigParser.ConfigParser()
  config.read("/etc/groovix/pacauth.conf")
 
  try:
    sipserver=config.get('pacauth','sipserver')
  except:
    sipserver=""
  if (not sipserver):
    return (34,"sip authentication method server not configured")

  sipargs=config.get('pacauth','sipargs')
 
  execlist=["sip_command.pl","--server",sipserver,"--command=patroninfo","--patron_id",mysipname,"--patron_pw",mypassword]
  #execlist+=sipargs.split()
  execlist+=shlex.split(sipargs)
  myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
  #this will block until command exits
  rm=myprocess.communicate()[0]
  rc=myprocess.returncode
  return (rc,rm)



###################################################
# III web based authentication
###################################################

def iii_auth(myusername,passwd):

  debug=0

  #test username
  match=re.search(r'^u(\d+\-*\d+)$',myusername)
  #unless we have compatible name don't even try custom auth
  if (not match):
    return (33,"iii authentication not applicable for this username")
  username=match.group(1)

  config=ConfigParser.ConfigParser()
  config.read("/etc/groovix/pacauth.conf")

  try:
     iii_server=config.get('pacauth','iii_server')
  except:
     iii_server=""
  if (not iii_server):
    return (34,"iii authentication method server not configured")

  try:
     workstation_type=config.get('pacauth','workstation_type')
  except:
     workstation_type="default"

  try:

    if re.match('^\d+$',username) and re.match('^\d+$',passwd):
        pass
    else:
        return 9,"Please enter your barcode and pin without any spaces or other characters"

    h = Http()

    #check that barcode and pin are numbers only - no need to urlencode anything

    url='http://'+iii_server+'/PATRONAPI/'+username+'/'+passwd+'/pintest'

    resp, content = h.request(url);
    dcontent=content.decode('utf-8');
    if debug:
        print resp
        print dcontent

    if (content.find('ERRNUM' ) > -1):
           rc=1
           rm="invalid barcode or pin"
    else:
           rc=0
           rm="OK"


    if rc==0:
         url='http://'+iii_server+'/PATRONAPI/'+username+'/dump'
         resp, content = h.request(url);
         dcontent=content.decode('utf-8');
         if debug:
             print resp
             print dcontent
         pcode1=re.search('\[p44\]=([^<]*)<BR>',dcontent);
         if pcode1:
             if debug:
                 print pcode1.group(1)
             if pcode1.group(1)=='1' :
                 if workstation_type!="child":
                     #this is an adult workstation and a child is trying to logon!
                     rc=3
                     rm="this workstation is for adults only"
             elif pcode1.group(1)=='0' :
                     rc=4
                     rm="internet access is not allowed for this account"

         #expdate=re.search('\[p43\]=([^<]*)<BR>',dcontent);
         expdate=re.search('\[p43\]=(\d\d)-(\d\d)-(\d\d)<BR>',dcontent);
         if expdate:

             #syslog.syslog("full output: "+str(dcontent))
             #syslog.syslog("expdate raw: "+str(expdate))
             #syslog.syslog("now time: "+str(datetime.now()))
             #syslog.syslog("expdate raw group 1: "+str(expdate.group(1)))
             #syslog.syslog("expdate from string : "+str(datetime.strptime("11-12-13","%m-%d-%y")))
             #syslog.syslog("now time: "+str(datetime.now()))
             #syslog.syslog("expdate: "+str(datetime.strptime(expdate.group(1),"%m-%d-%y")))

             expnum=(int(expdate.group(1))*100)+int(expdate.group(2))+((int(expdate.group(3))+2000)*10000)
             nownum=(int(datetime.now().strftime('%m'))*100)   +(int(datetime.now().strftime('%d')))   +((int(datetime.now().strftime('%y'))+2000)*10000)

             #if debug:
             #log this for a while to make sure its working: strptime was failing intermittently - only when called through gdm through c - with this exception:
             #('TypeError', '<no args>', TypeError("attribute of type 'NoneType' is not callable",), ['  File "/usr/share/pacauth/src/custom_auth.py", line 136, in iii_auth\n    syslog.syslog("expdate: "+str(datetime.datetime.strptime(expdate.group(1),"%m-%d-%y")))\n'])

             if True:
                 syslog.syslog("exp num: "+str(expnum))
                 syslog.syslog("now num: "+str(nownum))

             #if datetime.strptime(expdate.group(1),"%m-%d-%y") < datetime.now() :
             if expnum < nownum :
                 rc=5
                 rm="this account has expired"

    #if rc!=0:
    #    #by customer request: one failure message for almost everything!
    #    rm="ACCESS DENIED: If your login is unsucessful, please go to the circulation desk."

  except:
    rc=2
    rm="could not communicate with iii authentication server, allowing login despite exception: "+str(formatExceptionInfo())

  if debug:
     print "iii_auth rc =",rc,"rm =",rm

  return (rc,rm)


#########################################
# XML
#########################################

def xml_auth(username,mypassword):

  #specific to albany - the only one using this 
  match=re.search(r'^([PpCc][Aa][Ll][Bb].*)$',username)
  if (not match):
    return (33,"xml authentication not applicable for this username")
  myxmlname=match.group(1)

  config=ConfigParser.ConfigParser()
  config.read("/etc/groovix/pacauth.conf")

  try:
    xmlserver=config.get('pacauth','xmlserver')
  except:
    xmlserver=""
  if (not xmlserver):
    return (34,"xml authentication method server not configured")
  xmlpath=config.get('pacauth','xmlpath')

  execlist=["xml_authenticator.pl","--server",xmlserver,"--serverpath",xmlpath,"--patron_id",myxmlname,"--patron_pw",mypassword]

  myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
  #this will block until command exits
  rm=myprocess.communicate()[0]
  rc=myprocess.returncode
  return (rc,rm)

def formatExceptionInfo(maxTBlevel=4):
         import traceback
         import sys
         cla, exc, trbk = sys.exc_info()
         excName = cla.__name__
         try:
             excArgs = exc.__dict__["args"]
         except KeyError:
             excArgs = "<no args>"
         excTb = traceback.format_tb(trbk, maxTBlevel)
         return (excName, excArgs, exc, excTb)


