#!/usr/bin/env python

#######################################
# Groovix PacAuth
#######################################

########################################################################
# Copyright 2007-2011  Open Sense Solutions LLC  http://open-sense.com
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.  You should have received a copy of the GNU General Public License along with this program.  If not, see http://www.gnu.org/licenses/
########################################################################



import os
import fnmatch
import commands
import sys
import re

import syslog
import ConfigParser
import subprocess
import xmlrpclib
import spwd, crypt
import pwd


import custom_auth

# in python version < 2.7, popen clobbers the signal handler
#  passed from gdm, so we need to save/restore it in the pam c module
#  Popen needs the default SIGCHLD handler in place to work properly or you
#  intermittently get a "no child processes" exception
import signal
signal.signal(signal.SIGCHLD, signal.SIG_DFL)


#from optparse import OptionParser



#########################################
#FUNCTIONS
#########################################

def formatExceptionInfo(maxTBlevel=4):
         import traceback
         import sys
         cla, exc, trbk = sys.exc_info()
         excName = cla.__name__
         try:
             excArgs = exc.__dict__["args"]
         except KeyError:
             excArgs = "<no args>"
         excTb = traceback.format_tb(trbk, maxTBlevel)
         return (excName, excArgs, exc, excTb)



def check_for_local_login(user):
 global mypid
 global logincheckpath
 global myunixname
 rc=0

 try:
  #what if they are ABOUT to be logged in?
  #two checks, originally set to 120 and 2 seconds, corresponding to the beginning or end of pacauth
  #1.  the 120 sec. marker is written at the beginning of pacauth and should be removed when pacauth exits
  #  it prevents another pacauth from letting the user in while this process is running, or if it exits abnormally
  #     it prevents another login with 120 from when the other process started
  #2.  the 2 sec. marker is written at the end, to handle the time between the pacauth exits until the users is actually
  # logged in and they show up in the output of `users`.  
  # if there is another processes end marker less than 2 seconds old, we've in that between time and we have to disallow the login

  mystartfile=logincheckpath+myunixname+'.'+str(beginlimit)+"."+str(mypid)
  myendfile=  logincheckpath+myunixname+'.'+str(endlimit)+"."+str(mypid)
  #update marker for this process
  #gsorc,output=commands.getstatusoutput('touch '+mystartfile)
  open(mystartfile, 'a').close() 

  for curtype in (beginlimit,endlimit):
   for otherfile in findfiles(myunixname+'.'+str(curtype)+'.*',logincheckpath):
    if (otherfile == mystartfile):
     #this is ourself, ignore
     pass
    else:
     #another login process is running (or recently ran) for this user
     ourtime=os.path.getmtime(mystartfile)
     othertime=os.path.getmtime(otherfile)
     difftime=ourtime - othertime
     if ( ( difftime > curtype ) or ( difftime < 0 ) ):
         #otherfile is old or in the future, get rid of it to remove cruft
         os.remove(otherfile)
     else:
       #not OK, there is another login process active for this user!
       rc=12
       rm='login attempted less than '+str(curtype)+' seconds ago'
 

  if (rc != 0):
    return (rc,rm)

  #now do easier check to see if they are already logged in
  execlist=["users"]
  myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE)

  #this will block until command exits, seems to be more problematic than readline/wait
  #returnlist=myprocess.communicate()[0]
  #this works, but I don't know why it should be better than communicate!
  #readline will grab one line (waiting for a newline)
  returnlist=myprocess.stdout.readline()
  #make sure the process ends to avoid zombies
  myprocess.wait()

  userlist=returnlist.split()
  rc=myprocess.returncode
  if (userlist.count(myunixname)):
    rm="user already logged in!"
    rc=9
  else:
    rm="user not already logged in"
    rc=0 

 except:
    #debug# rm=str(formatExceptionInfo())
    #rm=str(formatExceptionInfo())
    rm="another login process already active for this user!"
    rc=19

 return (rc,rm)
##########################################
# easier way for local login dup check
#   found out about the getstatusoutput command after the fact
#########################################
#    import commands
#    gsorc,users=commands.getstatusoutput('users')
#    for localuser in localusers.rsplit():
#      if (localuser == myunixname):
#        rc=9
#        rm="local user already logged in!"


#not used anymore, disabled
def check_for_remote_login(remotescript,myversion,mypass,timezone,user,machine,display):
  global myunixname
  #make sure not logged in within 90 seconds somewhere else
  #seconds=remotescript.preventDupLogins(myversion,mypass,timezone,myunixname,machine,display)
  #with new pactime, duplicate logins will be dealt with while session is running, 
  # always allow login initially to avoid hassling users whose machine just crashed
  # TODO: take this code/check completely out
  seconds=91
  #print "seconds is "+str(seconds)
  if (seconds < 90):
    rm="already logged in less than 90 seconds ago"
    rc=90
  else:
    rm="not already logged in somewhere else"
    rc=0
  return (rc,rm)

def xmlauth(myxmlname,mypassword):
   
  config=ConfigParser.ConfigParser()
  config.read("/etc/groovix/pacauth.conf")
  
  xmlserver=config.get('pacauth','xmlserver')
  xmlpath=config.get('pacauth','xmlpath')
  
  execlist=["xml_authenticator.pl","--server",xmlserver,"--serverpath",xmlpath,"--patron_id",myxmlname,"--patron_pw",mypassword]

  myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
  #this will block until command exits
  rm=myprocess.communicate()[0]
  rc=myprocess.returncode
  return (rc,rm)


def local_auth(myunixname,mypassword):
  try:
    pw1 = spwd.getspnam(myunixname).sp_pwd
    pw2 = crypt.crypt(mypassword, pw1)
    if ( pw1 == pw2 ):
      rc=0
      rm="Local authentication successful."
    else:
      rc=6
      rm="Local username/password is invalid."
  except:
    #getspnam throws exception if user doesn't exist
    rc=7
    rm="Local user denied."
  return (rc,rm)

def finish_up(rc,rm,myunixname,mypassword,fullLoginPrep):

  global mypid
  global logincheckpath
  mystartfile=logincheckpath+myunixname+'.'+str(beginlimit)+"."+str(mypid)
  myendfile=  logincheckpath+myunixname+'.'+str(endlimit)+"."+str(mypid)
  #for debugging
  #import time
  #time.sleep(5)
  try:
    gsorc,output=commands.getstatusoutput('mv '+mystartfile+' '+myendfile)
  except:
    pass

  if (rc != 0) and (rc != 2):
    #print "Error: "+ rm + ", code "+str(rc)
    returnMessage= "Error: "+ rm 
  else:
    #print "Success: "+ rm + ", code "+str(rc)
    returnMessage= "Success: "+ rm 


    if (fullLoginPrep):
      #auth was OK
      #lets add the user if they don't exist or don't have the home dir populated
      #os.system("id -u "+myunixname+" 2>&1 > /dev/null || groovix-kiosk-create-user "+myunixname+" 2>&1 > /dev/null")
  
      execlist=['id','-u',myunixname]
      myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
      #this will block until command exits
      returnout,returnerr=myprocess.communicate()
      idrc=myprocess.returncode
      #print "idrc is "+str(idrc)+" out is "+returnout+ " err is "+returnerr
      if (idrc != 0):
        #print "no user by this name yet, let's create"
        #NOTE: if you manually add a local user, you also need to manually run groovix-kiosk-create-user
        execlist=['groovix-create-user',myunixname]
        myprocess=subprocess.Popen(execlist,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        #this will block until command exits
        returnout,returnerr=myprocess.communicate()
        curc=myprocess.returncode
        #print "curc is "+str(curc)+" out is "+returnout+ " err is "+returnerr
  
        #new user created, let's set local password for other non-pam friendly things to authenticate with (stupid xlock)
        #    import commands
        cucommand='usermod --password `echo "'+mypassword+'" | mkpasswd -s -H MD5 `'+' '+myunixname
        curc,output=commands.getstatusoutput(cucommand)
        #print "result of change password command: "+output


  #print 'pacauth result for user '+myunixname+' is '+str(rc)+' :  '+rm
  syslog.syslog('pacauth result for user '+myunixname+' is '+str(rc)+' :  '+rm);

  return rc,returnMessage


def findfiles(pattern, root='/var/lib/groovix/pacauth/logins'):
    #for recursively finding: # for path, dirs, files in os.walk(os.path.abspath(root)):
        files = os.listdir(os.path.abspath(root))
        for filename in fnmatch.filter(files, pattern):
          yield os.path.join(root, filename)

#########################################
#MAIN 
#########################################

def run(myusername,mypassword,fullLoginPrep):
 #fullLoginPrep as opposed to just a simple auth like we use for pacQ to know if they are a valid user
 global logincheckpath
 global beginlimit
 global endlimit
 global mypid
 global myunixname
 mypid=os.getpid()
 logincheckpath='/var/lib/groovix/pacauth/logins/'
 #define time limits
 beginlimit=120
 endlimit=2

 #########################################
 #get args
 #########################################

 myversion=None;
 mypass=None;
 timezone=None;
 enablePacQ=None;
 enablePacRez=None;
 machine=None;
 pacauthserver=None;

 #we only need to know the machine name for fullloginprep (and then it is only used for pacrez/pacQ)
 if (fullLoginPrep):
   if (machine == None):
     #f=open('/etc/groovix/id', 'r')
     #machine=(f.read()).strip()
     machine=(subprocess.check_output(["groovix-get-parameter","GX_LABEL"])).strip()


 #if not defined already, get from config file
 config=ConfigParser.ConfigParser()
 config.read("/etc/groovix/pacauth.conf")
 if (myversion == None):
  myversion=config.get('pacauth','customid')
 if (mypass == None):
  mypass=config.get('pacauth','custompw')
 if (timezone == None):
  timezone=config.get('pacauth','timezone')
 if (enablePacQ == None):
  try:
    enablePacQ=int(config.get('pacauth','enablepacq'))
  except:
    enablePacQ=None
 if (enablePacRez == None):
  try:
    enablePacRez=int(config.get('pacauth','enablepacrez'))
  except:
    enablePacRez=None
 if (pacauthserver == None):
  pacauthserver=config.get('pacauth','pacauthserver')

 #check if localauthonly is set in config file
 try:
    localauthonly=int(config.get('pacauth','localauthonly'))
 except:
    localauthonly=False


 myunixname=myusername

 try:
  display=os.environ['DISPLAY']
 except:
  #if not set, assume display 0, useful testing pacq and others
  display="0"
 match=re.search(r'\:*(\d)\.*(\d)*',display)
 if match:
  display=match.group(1)
  
 #default to error state
 rc=-2
 rm="error"

 if (fullLoginPrep):


  #########################################
  # QUICK CHECK FOR PAC Queue (SIMPLE-Q) RESERVED STATE before we bother authenticating
  #########################################
  rc=0
  rezuser=""
  if (enablePacQ):
    try:
     sys.path.append('/usr/share/pacq/src')
     from pacq import pacqRezCheck 
     (rc,rezuser,rm)=pacqRezCheck(None,None,None,None,machine,display)
    #handle anything including timeouts
    # except xmlrpclib.ProtocolError:
    except:
     rm="Can't connect to PAC Queue access server " 
     rc=14
    
    #for pac-q, 0 doesn't equate to "success" and 1 doesn't equate to "failure"
    # rc=0 means there is no reservation (or couldn't connect to server), which means the machine
    # is open and anyone can login.  
    # rc=1 means there is a reservation which requires further investigation
    if (rc == 1):
     if (rezuser == myunixname):
       rc=0
       rm="reservation is for the correct user"
     else:
       rc=15
       rm="this machine is reserved for user "+rezuser
       return finish_up(rc,rm,myunixname,mypassword,fullLoginPrep)


  #########################################
  # CHECK LOCAL DUPLICATE LOGINS
  #########################################
  (rc,rm)=check_for_local_login(myunixname)
  if (rc != 0):
   return finish_up(rc,rm,myunixname,mypassword,fullLoginPrep)
 
  #set to unsuccessful in case custom auth is skipped
  rc=1


 #end if full login prep
 
 #########################################
 # first try custom auth, if it is not applicable,
 # try guest.* : we assume guest login
 # all other user names : local auth
 #########################################

 # new for 2013: first try passwordless login
 try:
   pwlfile = open("/etc//groovix/passwordlesslogins", "r")
   for line in pwlfile:
    #print line.rstrip()
    #print myusername
    if ( line.rstrip() == myusername ):
       # we have a match .  but if there is a .xlockpw file in the home directory we need to heed that
       try:
          # if this file exists then we must match it
          userinfo=pwd.getpwnam(myusername)
          userhome=userinfo.pw_dir
          xlpwfile = open(userhome+"/.xlockpw", "r")
          if ( (xlpwfile.read()).rstrip() == mypassword ):
                     rc=0
                     rm="passwordless xlock pin matched"
                     syslog.syslog('pacauth passwordless xlock pin match result for user '+myusername+' is '+str(rc)+' :  '+rm);
          else:
                     rc=25
                     rm="passwordless xlock pin not matched"
                     syslog.syslog('pacauth passwordless xlock pin match result for user '+myusername+' is '+str(rc)+' :  '+rm);
          break
       except:
          # no .xlockpw file exists, so auth is success
          pass
          rc=0
          #print "match"
          rm="passwordless login matched"
          syslog.syslog('pacauth passwordless result for user '+myusername+' is '+str(rc)+' :  '+rm);
          break
 except:
   #print "exception"
   rc=33;


 #if it was a successful passwordless login, can skip the rest
 if ( rc != 0 ):

 #########################################
 # custom auth
 #########################################
  if (localauthonly):
    #custom method is not applicable
    rc=33; 
  else:
    (rc,rm)=custom_auth.my_auth(myusername,mypassword)
    syslog.syslog('pacauth custom result for user '+myusername+' is '+str(rc)+' :  '+rm);

     #########################################
     # guest auth
     #########################################
    if ( rc == 33 or rc == 34):
       #custom auth method was not applicable, try guest auth 
       # match=re.search(r'^guest',myusername)
       match=re.search(r'^g(uest)*\d',myusername)
       #unless we have a guest username don't even try
       if ( not match ):
         #guest auth is not applicable either
         rc=33;
       else:
         #let's try guest auth
         #########################################
         # SET UP XML RPC
         # used only for guest passes now
         #########################################
         try:
           remotescript = xmlrpclib.ServerProxy(pacauthserver)
         #handle anything including timeouts
         except:
           #old:# except xmlrpclib.ProtocolError: 
           rm="Can't connect to user access server at " + pacauthserver
           rc=5
         try:
           (rc,rm)=remotescript.guestAuth(myversion,mypass,timezone,myunixname,mypassword,fullLoginPrep)
         except:
           rm="guest authentication unavailable from " + pacauthserver
           rc=8

 if ( rc == 33 or rc == 34):
         #both custom and guest methods were not applicable, try local auth
         (rc,rm)=local_auth(myunixname,mypassword) 
  
 #########################################
 # not used anymore - CHECK REMOTE DUPLICATE LOGINS
 #########################################
 # need even for basic pacQ (simpleq) auth to prevent someone already logged in from making a new reservation
 #if (rc == 0):
  #good up till this point
  #one more thing to test: check for duplicate remote login 
  #this also writes timestamp to prevent near-simultaneous logins

  #check_for_remote_login not used anymore
  #try:
  #    (rc,newrm)=check_for_remote_login(remotescript,myversion,mypass,timezone,myunixname,machine,display)
  #    if (rc != 0):
  #      #only if fail here do we change return message
  #      rm=newrm
  #  #except ZeroDivisionError:
  #  #change to allow exception through for debug
  #  except:
  #    #if something went wrong, forget about duplicate check and let them pass
  #    rc=0;
  #    rm="could not check for duplicate remote logins using server at " + pacauthserver
 
 if (fullLoginPrep):

  #########################################
  # NOW, CHECK FOR PAC REZ RESERVED STATE 
  #  (unlike pacQueue we must wait until after authentication for rez check because we may make a
  #  reservation in the process and only want to do that if the user is authenticated)
  #########################################

  if rc==0 or rc==2 :
   #if we've authenticated succesfully so far (or authentication connection method failed and we are allowing with rc==2)
   rezuser=""
   if (enablePacRez):
    try:
     sys.path.append('/usr/share/pacrez/src')
     from pacrez import pacRezCheck 
     #(rc,rezuser,rm)=pacRezCheck(None,None,None,None,None)
     (rc,rm,nowtime,crc,rezuser,c_start_time,c_end_time,frc,frezuser,f_start_time,f_end_time)=pacRezCheck(None,None,None,None,None)
    #handle anything including timeouts
    #except xmlrpclib.ProtocolError:
    except:
     rm="Can't connect to PAC Reservations access server " 
     rc=14
    
    
    #for pac-rez, main rc is a failure, we'll just ignore it for now
    # rc=0 means checking for a reservation was successful
    # crc=0 means there is a current reservation
    if (rc == 0):
     if (crc == 0):
      if (rezuser == myunixname):
       rc=0
       rm="reservation is for the correct user"
      else:
       rc=15
       #rm="this machine is reserved for another user until " + str((c_end_time.replace(second=59,microsecond=0)).time() )
       rm="this machine is reserved for another user until " + str(c_end_time.strftime("%I:%M %p")  )
       return finish_up(rc,rm,myunixname,mypassword,fullLoginPrep)
     else:
      #we need to make a reservation
      from pacrez import makeRez
      #rc,rm=makeRez(None,None,None,None,None,myunixname,"now","max") 
      #make the rez up until 1 minute before the next reservation
      if ( frc == 0 ) and ( frezuser == myunixname):
          #our reservation is coming up next, let them log in now and
          #keep the original duration the same, add one minute since it will be floored down by int
          #edit_entry_handler will sense the conflict and edit the original reservation
          #proposed_duration=int((f_end_time-f_start_time).seconds/60)+1
          #now that we're dealing with the actual end time instead of end time - 30 seconds, this should be exactly right
          proposed_duration=int((f_end_time-f_start_time).seconds/60)
      else:
         #we have no future reservation or a different user is up next
         if ( frc != 0 ):
          #no future reservation
          proposed_duration="max";
         else:
           # a different user is up next
           #set duration to next reservation start - now - 1 minute
           proposed_duration=int((f_start_time-nowtime).seconds/60)-1

      #print f_start_time
      #print nowtime
      #print proposed_duration
      rc,rm=makeRez(None,None,None,None,None,myunixname,"now",proposed_duration ) 
      #print "debug: rc is",rc
      if (rc != 0) and (rc !=2 ):
         #unless success or unable to connect, fail
         return finish_up(rc,rm,myunixname,mypassword,fullLoginPrep)

  #print "debug: rc is",rc

  #########################################
  # DOUBLE CHECK LOCAL DUPLICATE LOGINS
  #  lots went on since we first checked, machine could crash if we allow local dups, so check again
  #########################################
  if (rc == 0) or (rc == 2):
   #if we've authenticated succesfully so far
   (lrc,lrm)=check_for_local_login(myunixname)
   #we could use a semaphore to really prevent race conditions
   # only if there is a problem, change rc and rm
   if (lrc != 0):
    rc=lrc
    rm=lrm

 #########################################
 # FINISH UP
 #########################################
 return finish_up(rc,rm,myunixname,mypassword,fullLoginPrep)

